package jess;

import java.io.Serializable;
import java.util.*;

/**
 * An Agenda represents the ordered lists of activated rules for all
 * modules defined in a single Rete engine.
 * <p/>
 * (C) 2006 Sandia National Laboratories<br>
 */

class Agenda implements Serializable {

    private final Object m_activationSemaphore = new String("ACTIVATION LOCK");
    private volatile boolean m_halt = false;
    private int m_evalSalience = Rete.INSTALL;
    private HashMap m_modules = new HashMap();
    private String m_thisModule;
    private Strategy m_strategy = new depth();
    private final Stack m_focusStack = new Stack();
    private Activation m_thisActivation;

    Agenda(Rete engine) {
        try {
            addDefmodule(new Defmodule(), engine);
        } catch (JessException je) {
            throw new RuntimeException("Can't define module MAIN");
        }
    }

    void setEvalSalience(int method) throws JessException {
        if (method < Rete.INSTALL || method > Rete.EVERY_TIME)
            throw new JessException("Agenda.setEvalSalience",
                    "Invalid value", method);
        m_evalSalience = method;
    }

    int getEvalSalience() {
        return m_evalSalience;
    }

    void reset(Rete engine) throws JessException {
        for (Iterator modules = m_modules.values().iterator();
             modules.hasNext();) {
            Defmodule module = (Defmodule) modules.next();
            module.reset();
        }
        synchronized (m_focusStack) {
            m_focusStack.clear();
            m_focusStack.push(Defmodule.MAIN);
        }
        engine.broadcastEvent(JessEvent.FOCUS, Defmodule.MAIN, null);
    }

    void clear(Rete engine) throws JessException {
        m_modules.clear();
        m_strategy = new depth();
        addDefmodule(new Defmodule(), engine);
        m_focusStack.clear();
        synchronized (m_activationSemaphore) {
            m_halt = false;
            m_activationSemaphore.notifyAll();
        }
    }

    Object getActivationSemaphore() {
        return m_activationSemaphore;
    }

    HeapPriorityQueue getQueue() {
        return ((Defmodule) m_modules.get(m_thisModule)).getQueue();
    }

    HeapPriorityQueue getQueue(Object module) throws JessException {
        verifyModule(module);
        return ((Defmodule) m_modules.get(module)).getQueue();
    }

    Activation getNextActivation(Rete engine) throws JessException {
        synchronized (m_activationSemaphore) {
            synchronized (m_focusStack) {
                if (m_focusStack.empty())
                    return getQueue(Defmodule.MAIN).pop();

                while (!m_focusStack.empty()) {
                    HeapPriorityQueue q = getQueue(m_focusStack.peek());
                    Activation a = q.pop();
                    if (a != null)
                        return a;
                    else {
                        Object oldFocus = m_focusStack.pop();
                        engine.broadcastEvent(JessEvent.FOCUS | JessEvent.REMOVED,
                                oldFocus, null);
                        if (!getFocus().equals(oldFocus))
                            engine.broadcastEvent(JessEvent.FOCUS, getFocus(), null);
                    }
                }
                return getQueue(Defmodule.MAIN).pop();
            }
        }
    }

    Activation peekNextActivation(Rete engine) throws JessException {
        synchronized (m_activationSemaphore) {
            synchronized (m_focusStack) {
                if (m_focusStack.empty())
                    return getQueue(Defmodule.MAIN).peek();

                while (!m_focusStack.empty()) {
                    HeapPriorityQueue q = getQueue(m_focusStack.peek());
                    Activation a = q.peek();
                    if (a != null)
                        return a;
                    else {
                        Object oldFocus = m_focusStack.pop();
                        engine.broadcastEvent(JessEvent.FOCUS | JessEvent.REMOVED,
                                oldFocus, null);
                        if (!getFocus().equals(oldFocus))
                            engine.broadcastEvent(JessEvent.FOCUS, getFocus(), null);
                    }
                }
                return getQueue(Defmodule.MAIN).peek();
            }
        }
    }


    Iterator listActivations() {
        return getQueue().iterator();
    }

    Iterator listActivations(String moduleName) throws JessException {
        return getQueue(moduleName).iterator();
    }

    Iterator listModules() {
        return m_modules.keySet().iterator();
    }


    private HashSet m_toAdd = new HashSet();
    private HashSet m_toRemove = new HashSet();

    void addActivation(Activation a) {
        synchronized (m_activationSemaphore) {
            m_toAdd.add(a);
        }
    }

    void removeActivation(Activation a) {
        synchronized (m_activationSemaphore) {
            if (a.getIndex() == -1) {
                m_toAdd.remove(a);
            } else {
                m_toRemove.add(a);
            }
        }
    }

    void commitActivations(Rete engine) throws JessException {
        synchronized (m_activationSemaphore) {
            if (m_toAdd.size() == 0 && m_toRemove.size() == 0)
                return;

            try {
                for (Iterator it = m_toRemove.iterator(); it.hasNext();) {
                    Activation a = (Activation) it.next();
                    if (a.getIndex() > -1)
                        getQueue(a.getModule()).remove(a);
                }

                for (Iterator it = m_toAdd.iterator(); it.hasNext();) {
                    Activation a = (Activation) it.next();

                    if (m_evalSalience != Rete.INSTALL)
                        a.evalSalience(engine);

                    getQueue(a.getModule()).push(a);

                    if (a.getAutoFocus())
                        setFocus(a.getModule(), engine);

                }

            } finally {
                if (m_toAdd.size() > 0)
                    m_activationSemaphore.notifyAll();

                m_toAdd.clear();
                m_toRemove.clear();
            }
        }
    }

    void waitForActivations() {
        try {
            synchronized (m_activationSemaphore) {
                if (getQueue().isEmpty())
                    m_activationSemaphore.wait();
            }
        } catch (InterruptedException ie) { /* FALL THROUGH */
        }
    }

    Strategy getStrategy() {
        return m_strategy;
    }

    String setStrategy(Strategy s, Rete r) throws JessException {
        String rv = null;
        try {
            for (Iterator modules = m_modules.values().iterator(); modules.hasNext();) {
                Defmodule module = (Defmodule) modules.next();
                rv = module.setStrategy(s, r);
            }
        } finally {
            r.commitActivations();
        }
        m_strategy = s;
        return rv;
    }

    void halt() {
        synchronized (m_activationSemaphore) {
            m_halt = true;
            m_activationSemaphore.notifyAll();
        }
    }

    public boolean isHalted() {
        return m_halt;
    }

    int runUntilHalt(Rete r, Context context) throws JessException {
        int count = 0;
        do {
            count += run(r, context);
            synchronized (m_activationSemaphore) {
                if (m_halt)
                    break;
                waitForActivations();
                if (m_halt)
                    break;
            }

        } while (true);

        return count;
    }

    synchronized int run(Rete r, Context context) throws JessException {
        int i = 0, j;
        // ###
        do {
            j = run(Integer.MAX_VALUE, r, context);
            i += j;
            synchronized (m_activationSemaphore) {
                if (m_halt)
                    break;
            }

        } while (j > 0);
        return i;
    }

    synchronized int run(int max, Rete r, Context context) throws JessException {
        int n = 0;
        m_halt = false;
        Activation a;
        // ###

        while (n < max && (a = getNextActivation(r)) != null) {
            if (!a.isInactive()) {
                a.setSequenceNumber(++n);
                r.broadcastEvent(JessEvent.DEFRULE_FIRED, a, null);
                try {
                    r.aboutToFire(a);
                    m_thisActivation = a;
                    a.fire(r, context);
                } finally {
                    m_thisActivation = null;
                    r.justFired(a);
                }

                if (m_evalSalience == Rete.EVERY_TIME) {
                    setStrategy(getQueue().getStrategy(), r);
                }
            }

            synchronized (m_activationSemaphore) {
                if (m_halt)
                    break;
            }
        }
        return n;
    }

    String getCurrentModule() {
        return m_thisModule;
    }

    Defmodule getModule(String name) throws JessException {
        verifyModule(name);
        return (Defmodule) m_modules.get(name);
    }

    Iterator listFocusStack() {
        return m_focusStack.iterator();
    }

    void clearFocusStack() {
        m_focusStack.clear();
    }

    String getFocus() {
        synchronized (m_focusStack) {
            if (m_focusStack.empty())
                return Defmodule.MAIN;
            else
                return (String) m_focusStack.peek();
        }
    }

    String popFocus(Rete engine, String expect) throws JessException {
        synchronized (m_focusStack) {
            if (m_focusStack.empty())
                return Defmodule.MAIN;
            else if (expect != null && !expect.equals(getFocus()))
                return expect;
            else {
                String oldFocus = (String) m_focusStack.pop();

                engine.broadcastEvent(JessEvent.FOCUS | JessEvent.REMOVED,
                        oldFocus, null);
                engine.broadcastEvent(JessEvent.FOCUS, getFocus(), null);
                return oldFocus;
            }
        }
    }

    void setFocus(String name, Rete engine) throws JessException {
        if (getFocus().equals(name))
            return;
        verifyModule(name);
        engine.broadcastEvent(JessEvent.FOCUS | JessEvent.REMOVED, getFocus(), null);
        engine.broadcastEvent(JessEvent.FOCUS, name, null);
        m_focusStack.push(name);
    }

    String setCurrentModule(String moduleName) throws JessException {
        String old = m_thisModule;
        verifyModule(moduleName);
        m_thisModule = moduleName;
        return old;
    }

    void addDefmodule(Defmodule module, Rete engine) throws JessException {
        if (m_modules.get(module.getName()) != null)
            throw new JessException("Agenda.addDefmodule",
                    "Attempt to redefine defmodule",
                    module.getName());

        m_thisModule = module.getName();
        m_modules.put(module.getName(), module);
        module.setStrategy(m_strategy, engine);
    }

    void verifyModule(Object moduleName) throws JessException {
        if (m_modules.get(moduleName) == null)
            throw new JessException("Agenda.verifyModule",
                    "Undefined module",
                    moduleName.toString());
    }

    String resolveName(String name) {
        if (name.indexOf("::") == -1)
            name = RU.scopeName(getCurrentModule(), name);
        return name;
    }

    Activation getThisActivation() {
        return m_thisActivation;
    }
}
