package jess;

import java.io.Serializable;
import java.util.Iterator;
import java.util.TreeMap;

/**
 * A named module containing rules, templates, and other constructs.
 * <P>
 * (C) 2006 Sandia National Laboratories<BR>
 */


public class Defmodule implements Serializable, Named, Visitable {
    public static final String MAIN = "MAIN";
    private HeapPriorityQueue m_queue;
    private TreeMap m_deftemplates;
    private String m_name;
    private String m_comment;

    /**
     * Construct a module named MAIN.
     */
    public Defmodule() {
        this(MAIN, "Default module");
    }

    /**
     * Construct a new module.
     * @param name the name of the new module
     * @param comment documentation for the module, or null.
     */
    public Defmodule(String name, String comment) {
        m_name = name;
        m_comment = comment;
        m_deftemplates = new TreeMap();
    }

    String setStrategy(Strategy s, Rete engine) throws JessException {
        if (m_queue == null) {
            m_queue = new HeapPriorityQueue(s);
            return s.getName();
        }
        else {
            synchronized (m_queue) {
                Iterator e = m_queue.iterator();
                m_queue.clear();
                String name = m_queue.setStrategy(s).getName();
                while (e.hasNext()) {
                    Activation a = (Activation) e.next();
                    if (!a.isInactive())
                        engine.addActivation(a);
                }
                return name;
            }
        }
    }

    HeapPriorityQueue getQueue() {
        return m_queue;
    }

    /**
     * Returns the name of this module
     * @return the name
     */
    public String getName() {
        return m_name;
    }

    /**
     * Returns the docmentation string for this module.
     * @return the documentation string, or null if none.
     */
    public String getDocstring() {
        return m_comment;
    }

    void reset() {
        m_queue.clear();
    }

    /**
     * Return an iterator over all the deftemplates defined in this module.
     * @return the iterator
     */
    public Iterator listDeftemplates() {
        return m_deftemplates.values().iterator();
    }

    /**
     * Return a deftemplate defined in this module, by name.
     * @param name the name of the desired deftemplate
     * @return the deftemplate if found, or null
     */
    public Deftemplate getDeftemplate(String name) {
        int index = name.indexOf("::");
        if (index != -1) {
            name = name.substring(index+2);
        }
        return (Deftemplate) m_deftemplates.get(name);
    }

    /**
     * Add a new deftemplate to this moduole. The deftemplate must
     * already have been created in this module; it must already report
     * this module's name as part of its name. You can redefine a template
     * only if the definition is identical to the old one.
     * @param dt a new deftemplate
     * @param engine the Rete engine this module belongs to
     * @return the new deftemplate, or an existing identical template
     * @throws JessException if anything goes wrong.
     */
    public Deftemplate addDeftemplate(Deftemplate dt, Rete engine)
        throws JessException {
        synchronized (m_deftemplates) {
            if (!dt.getModule().equals(m_name))
                throw new JessException("Defmodule.addDeftemplate",
                                        "Wrong module name",
                                        dt.getModule());
            String name = dt.getBaseName();
            Deftemplate existing = (Deftemplate) m_deftemplates.get(name);
            if (existing == null || existing.equals(dt)) {
                engine.broadcastEvent(JessEvent.DEFTEMPLATE, dt, null);
                m_deftemplates.put(name, dt);
            } else {
                throw new JessException("Defmodule.addDeftemplate",
                                        "Cannot redefine deftemplate",
                                        dt.getName());
            }
            return dt;
        }
    }

    /**
     * Return the type of this construct.
     * @return the string "defmodule".
     */
    public final String getConstructType() {
        return "defmodule";
    }

    public Object accept(Visitor v) {
        return v.visitDefmodule(this);
    }

}
